/* git-annex-remote-clouda
   Copyright 2019 Timothy Sample <samplet@ngyro.com>

   This file is part of git-annex-remote-clouda.

   git-annex-remote-clouda is free software: you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   git-annex-remote-clouda is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with git-annex-remote-clouda.  If not, see
   <http://www.gnu.org/licenses/>. */

#include <curl/curl.h>
#include <libguile.h>
#include <stdio.h>
#include <stdlib.h>

void
curl_error (CURLcode error, const char *subr)
{
  scm_error (scm_from_utf8_symbol ("curl-error"),
             subr,
             curl_easy_strerror(error),
             SCM_EOL,
             scm_list_1 (scm_from_int (error)));
}

CURL *
curl_init (const char *subr)
{
  CURL *curl;

  curl = curl_easy_init ();
  if (curl == NULL) {
    scm_error (scm_from_utf8_symbol ("curl-error"),
               subr,
               "Could not initialize cURL",
               SCM_EOL,
               SCM_EOL);
  }

  scm_dynwind_unwind_handler (curl_easy_cleanup, curl,
                              SCM_F_WIND_EXPLICITLY);

  return curl;
}

void
set_url (CURL *curl, SCM url, const char *subr)
{
  char *url_c;
  CURLcode result;

  url_c = scm_to_utf8_stringn (url, NULL);
  scm_dynwind_free (url_c);

  result = curl_easy_setopt (curl, CURLOPT_URL, url_c);
  if (result != CURLE_OK) curl_error (result, subr);
}

struct curl_slist *
scm_headers_to_curl_slist (SCM list, const char *subr)
{
  struct curl_slist *slist = NULL, *tmp_slist = NULL;
  SCM list_head = list;
  char *header;

  while (scm_is_pair (list)) {
    header = scm_to_latin1_stringn (SCM_CAR (list), NULL);
    tmp_slist = curl_slist_append (slist, header);
    free (header);

    if (tmp_slist == NULL) {
      curl_slist_free_all (slist);
      scm_error (scm_from_utf8_symbol ("curl-error"),
                 subr,
                 "Could not append to curl_slist",
                 SCM_EOL,
                 SCM_EOL);
    }

    slist = tmp_slist;

    list = SCM_CDR (list);
  }

  if (!scm_is_eq (list, SCM_EOL)) {
    curl_slist_free_all (slist);
    SCM_ASSERT (0, list_head, SCM_ARGn, subr);
  }

  return slist;
}

void
add_headers (CURL *curl, SCM headers, const char *subr)
{
  struct curl_slist *headers_c;
  CURLcode result;

  headers_c = scm_headers_to_curl_slist (headers, subr);

  if (headers_c != NULL) {
    scm_dynwind_unwind_handler ((void (*)(void *))curl_slist_free_all,
                                headers_c,
                                SCM_F_WIND_EXPLICITLY);

    result = curl_easy_setopt (curl, CURLOPT_HTTPHEADER, headers_c);
    if (result != CURLE_OK) curl_error (result, subr);
  }
}

void
setup_ssl (CURL *curl, const char *subr)
{
  char *cert_dir, *cert_file;
  CURLcode result;

  cert_dir = getenv ("SSL_CERT_DIR");
  if (cert_dir != NULL) {
    result = curl_easy_setopt (curl, CURLOPT_CAPATH, cert_dir);
    if (result != CURLE_OK) curl_error (result, subr);
  }

  cert_file = getenv ("SSL_CERT_FILE");
  if (cert_file != NULL) {
    result = curl_easy_setopt (curl, CURLOPT_CAINFO, cert_file);
    if (result != CURLE_OK) curl_error (result, subr);
  }
}

/* This is here so that we can register a dynwind unwind handler
   that closes a file handle but also knows the name of the
   current subroutine in case it needs to raise an error. */
struct FILE_with_subr {
  FILE *file;
  const char *subr;
};

void
checked_fclose (struct FILE_with_subr *fws)
{
  if (fclose (fws->file) != 0) {
    scm_syserror (fws->subr);
  }
}

size_t
noop_write (char *ptr, size_t size, size_t nmemb, void *userdata)
{
  return size * nmemb;
}

SCM
curl_global_init_wrapper ()
{
  CURLcode result;

  if ((result = curl_global_init (CURL_GLOBAL_DEFAULT)) != CURLE_OK) {
    curl_error (result, "curl-global-init!");
  }

  return SCM_UNDEFINED;
}

SCM
curl_global_cleanup_wrapper ()
{
  curl_global_cleanup ();
  return SCM_UNDEFINED;
}

SCM
http_get_file (SCM url, SCM file, SCM headers)
{
  const char *subr = "%http-get-file";
  CURL *curl;
  char *file_c;
  FILE *out;
  struct FILE_with_subr fws;
  long response_code;
  CURLcode result;

  scm_dynwind_begin (0);

  curl = curl_init (subr);
  setup_ssl (curl, subr);
  set_url (curl, url, subr);
  add_headers (curl, headers, subr);

  file_c = scm_to_locale_string (file);
  scm_dynwind_free (file_c);

  out = fopen (file_c, "wb");
  if (out == NULL) scm_syserror (subr);
  fws.file = out;
  fws.subr = subr;
  scm_dynwind_unwind_handler ((void (*)(void *))checked_fclose, &fws,
                              SCM_F_WIND_EXPLICITLY);

  result = curl_easy_setopt (curl, CURLOPT_WRITEDATA, out);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_perform (curl);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &response_code);
  if (result != CURLE_OK) curl_error (result, subr);

  scm_dynwind_end ();

  return scm_from_long (response_code);
}

SCM
http_put_file (SCM url, SCM file, SCM headers)
{
  const char *subr = "%http-put-file";
  CURL *curl;
  char *file_c;
  FILE *in;
  struct FILE_with_subr fws;
  long response_code;
  CURLcode result;

  scm_dynwind_begin (0);

  curl = curl_init (subr);
  setup_ssl (curl, subr);
  set_url (curl, url, subr);
  add_headers (curl, headers, subr);

  /* The UPLOAD option tells cURL to do a PUT request. */
  result = curl_easy_setopt (curl, CURLOPT_UPLOAD, 1L);
  if (result != CURLE_OK) curl_error (result, subr);

  file_c = scm_to_locale_string (file);
  scm_dynwind_free (file_c);

  in = fopen (file_c, "rb");
  if (in == NULL) scm_syserror (subr);
  fws.file = in;
  fws.subr = subr;
  scm_dynwind_unwind_handler ((void (*)(void *))checked_fclose, &fws,
                              SCM_F_WIND_EXPLICITLY);

  result = curl_easy_setopt (curl, CURLOPT_READDATA, in);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_perform (curl);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &response_code);
  if (result != CURLE_OK) curl_error (result, subr);

  scm_dynwind_end ();

  return scm_from_long (response_code);
}

SCM
http_head (SCM url, SCM headers)
{
  const char *subr = "%http-head";
  CURL *curl;
  long response_code;
  CURLcode result;

  scm_dynwind_begin (0);

  curl = curl_init (subr);
  setup_ssl (curl, subr);
  set_url (curl, url, subr);
  add_headers (curl, headers, subr);

  /* The NOBODY option tells cURL to do a HEAD request. */
  result = curl_easy_setopt (curl, CURLOPT_NOBODY, 1L);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_perform (curl);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &response_code);
  if (result != CURLE_OK) curl_error (result, subr);

  scm_dynwind_end ();

  return scm_from_long (response_code);
}

SCM
http_delete (SCM url, SCM headers)
{
  const char *subr = "%http-delete";
  CURL *curl;
  long response_code;
  CURLcode result;

  scm_dynwind_begin (0);

  curl = curl_init (subr);
  setup_ssl (curl, subr);
  set_url (curl, url, subr);
  add_headers (curl, headers, subr);

  result = curl_easy_setopt (curl, CURLOPT_CUSTOMREQUEST, "DELETE");
  if (result != CURLE_OK) curl_error (result, subr);

  /* Withtout this, cURL writes the response body to standard
     out, breaking the git-annex protocol.  We have no need for
     the body, so we ignore it. */
  result = curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, noop_write);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_perform (curl);
  if (result != CURLE_OK) curl_error (result, subr);

  result = curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &response_code);
  if (result != CURLE_OK) curl_error (result, subr);

  scm_dynwind_end ();

  return scm_from_long (response_code);
}

void
git_annex_remote_clouda_init ()
{
  scm_c_define_gsubr ("curl-global-init!", 0, 0, 0,
                      (void (*)())curl_global_init_wrapper);
  scm_c_define_gsubr ("curl-global-cleanup!", 0, 0, 0,
                      (void (*)())curl_global_cleanup_wrapper);
  scm_c_define_gsubr ("%http-get-file", 3, 0, 0, (void (*)())http_get_file);
  scm_c_define_gsubr ("%http-put-file", 3, 0, 0, (void (*)())http_put_file);
  scm_c_define_gsubr ("%http-head", 2, 0, 0, (void (*)())http_head);
  scm_c_define_gsubr ("%http-delete", 2, 0, 0, (void (*)())http_delete);
}
